@@ -16,7 +16,7 @@ import subprocess, os.path, tempfile
 
 from .. import mlog
 from .. import coredata
-from ..mesonlib import EnvironmentException, version_compare, Popen_safe
+from ..mesonlib import EnvironmentException, version_compare, Popen_safe, listify
 
 from .compilers import (
     GCC_MINGW,
@@ -172,6 +172,9 @@ class CCompiler(Compiler):
     def get_linker_search_args(self, dirname):
         return ['-L' + dirname]
 
+    def get_default_include_dirs(self):
+        return []
+
     def gen_import_library_args(self, implibname):
         """
         The name of the outputted import library
@@ -482,6 +485,34 @@ class CCompiler(Compiler):
         # minus the extra newline at the end
         return p.stdo.split(delim + '\n')[-1][:-1]
 
+    def get_return_value(self, fname, rtype, prefix, env, extra_args, dependencies):
+        if rtype == 'string':
+            fmt = '%s'
+            cast = '(char*)'
+        elif rtype == 'int':
+            fmt = '%lli'
+            cast = '(long long int)'
+        else:
+            raise AssertionError('BUG: Unknown return type {!r}'.format(rtype))
+        fargs = {'prefix': prefix, 'f': fname, 'cast': cast, 'fmt': fmt}
+        code = '''{prefix}
+        #include <stdio.h>
+        int main(int argc, char *argv[]) {{
+            printf ("{fmt}", {cast} {f}());
+        }}'''.format(**fargs)
+        res = self.run(code, env, extra_args, dependencies)
+        if not res.compiled:
+            m = 'Could not get return value of {}()'
+            raise EnvironmentException(m.format(fname))
+        if rtype == 'string':
+            return res.stdout
+        elif rtype == 'int':
+            try:
+                return int(res.stdout.strip())
+            except:
+                m = 'Return value of {}() is not an int'
+                raise EnvironmentException(m.format(fname))
+
     @staticmethod
     def _no_prototype_templ():
         """
@@ -896,6 +927,9 @@ class VisualStudioCCompiler(CCompiler):
     def get_linker_search_args(self, dirname):
         return ['/LIBPATH:' + dirname]
 
+    def get_gui_app_args(self):
+        return ['/SUBSYSTEM:WINDOWS']
+
     def get_pic_args(self):
         return [] # PIC is handled by the loader on Windows
 
@@ -1013,8 +1047,7 @@ class VisualStudioCCompiler(CCompiler):
 
     def get_link_whole_for(self, args):
         # Only since VS2015
-        if not isinstance(args, list):
-            args = [args]
+        args = listify(args)
         return ['/WHOLEARCHIVE:' + x for x in args]
 
     def get_instruction_set_args(self, instruction_set):
@@ -1027,4 +1060,33 @@ class VisualStudioCCompiler(CCompiler):
             return None
         return vs32_instruction_set_args.get(instruction_set, None)
 
+    def get_toolset_version(self):
+        # See boost/config/compiler/visualc.cpp for up to date mapping
+        try:
+            version = int(''.join(self.version.split('.')[0:2]))
+        except:
+            return None
+        if version < 1310:
+            return '7.0'
+        elif version < 1400:
+            return '7.1' # (Visual Studio 2003)
+        elif version < 1500:
+            return '8.0' # (Visual Studio 2005)
+        elif version < 1600:
+            return '9.0' # (Visual Studio 2008)
+        elif version < 1700:
+            return '10.0' # (Visual Studio 2010)
+        elif version < 1800:
+            return '11.0' # (Visual Studio 2012)
+        elif version < 1900:
+            return '12.0' # (Visual Studio 2013)
+        elif version < 1910:
+            return '14.0' # (Visual Studio 2015)
+        elif version < 1920:
+            return '14.1' # (Visual Studio 2017)
+        return None
 
+    def get_default_include_dirs(self):
+        if 'INCLUDE' not in os.environ:
+            return []
+        return os.environ['INCLUDE'].split(os.pathsep)
