@@ -195,10 +195,10 @@ class TestHarness:
     def run_single_test(self, wrap, test):
         if test.fname[0].endswith('.jar'):
             cmd = ['java', '-jar'] + test.fname
-        elif not test.is_cross and run_with_mono(test.fname[0]):
+        elif not test.is_cross_built and run_with_mono(test.fname[0]):
             cmd = ['mono'] + test.fname
         else:
-            if test.is_cross:
+            if test.is_cross_built:
                 if test.exe_runner is None:
                     # Can not run test on cross compiled executable
                     # because there is no execute wrapper.
@@ -224,7 +224,7 @@ class TestHarness:
 
             child_env.update(test.env)
             if len(test.extra_paths) > 0:
-                child_env['PATH'] += os.pathsep.join([''] + test.extra_paths)
+                child_env['PATH'] = os.pathsep.join(test.extra_paths + ['']) + child_env['PATH']
 
             # If MALLOC_PERTURB_ is not set, or if it is set to an empty value,
             # (i.e., the test or the environment don't explicitly set it), set
