@@ -18,6 +18,7 @@ import shutil
 import argparse
 from .. import mlog
 from . import destdir_join
+from .gettext import read_linguas
 
 parser = argparse.ArgumentParser()
 parser.add_argument('command')
@@ -65,7 +66,7 @@ def install_help(srcdir, blddir, sources, media, langs, install_dir, destdir, pr
         for source in sources:
             infile = os.path.join(srcdir if lang == 'C' else blddir, lang, source)
             outfile = os.path.join(indir, source)
-            mlog.log('Installing %s to %s.' % (infile, outfile))
+            mlog.log('Installing %s to %s' % (infile, outfile))
             shutil.copyfile(infile, outfile)
             shutil.copystat(infile, outfile)
         for m in media:
@@ -74,14 +75,25 @@ def install_help(srcdir, blddir, sources, media, langs, install_dir, destdir, pr
             if not os.path.exists(infile):
                 if lang == 'C':
                     mlog.warning('Media file "%s" did not exist in C directory' % m)
+                    continue
                 elif symlinks:
                     srcfile = os.path.join(c_install_dir, m)
                     mlog.log('Symlinking %s to %s.' % (outfile, srcfile))
                     if '/' in m or '\\' in m:
                         os.makedirs(os.path.dirname(outfile), exist_ok=True)
-                    os.symlink(srcfile, outfile)
-                continue
-            mlog.log('Installing %s to %s.' % (infile, outfile))
+                    try:
+                        try:
+                            os.symlink(srcfile, outfile)
+                        except FileExistsError:
+                            os.remove(outfile)
+                            os.symlink(srcfile, outfile)
+                        continue
+                    except (NotImplementedError, OSError):
+                        mlog.warning('Symlinking not supported, falling back to copying')
+                else:
+                    # Lang doesn't have media file so copy it over 'C' one
+                    infile = os.path.join(srcdir, 'C', m)
+            mlog.log('Installing %s to %s' % (infile, outfile))
             if '/' in m or '\\' in m:
                 os.makedirs(os.path.dirname(outfile), exist_ok=True)
             shutil.copyfile(infile, outfile)
@@ -97,6 +109,9 @@ def run(args):
     build_subdir = os.path.join(os.environ['MESON_BUILD_ROOT'], options.subdir)
     abs_sources = [os.path.join(src_subdir, 'C', source) for source in sources]
 
+    if not langs:
+        langs = read_linguas(src_subdir)
+
     if options.command == 'pot':
         build_pot(src_subdir, options.project_id, sources)
     elif options.command == 'update-po':
